#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Virginia importer.
"""
import xlrd

from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.interfaces import IDemographics
from schooltool.course.interfaces import ICourseContainer, ISectionContainer
from schooltool.course.section import Section
from schooltool.group.interfaces import IGroupContainer
from schooltool.export import importer
from schooltool.skin import flourish
from schooltool.schoolyear.interfaces import ISchoolYearContainer

from schooltool.common import SchoolToolMessage as _


ERROR_INVALID_COURSE_ID = _("Provided Course ID was not found")
ERROR_INVALID_MARKING_PERIODS = _("Provided Marking Period IDs are not valid")
ERROR_INVALID_TEACHER_ID = _("Provided Teacher ID was not found")
ERROR_INVALID_STUDENT_ID = _("Provided Student ID was not found")
ERROR_INVALID_BUILDING_ID = _("Provided Building ID was not found")


class ImportSectionsLinkViewlet(flourish.page.LinkViewlet):

    @property
    def url(self):
        schoolyears = ISchoolYearContainer(self.context)
        schoolyear = schoolyears.getActiveSchoolYear()
        if 'schoolyear_id' in self.request:
            schoolyear_id = self.request['schoolyear_id']
            schoolyear = schoolyears.get(schoolyear_id, schoolyear)
        if schoolyear is None:
            return ''
        url = '%s/import-va-sections.html' % (
            absoluteURL(schoolyear, self.request))
        return url


class SectionsImporter(importer.ImporterBase):

    def import_data(self, wb):
        app = ISchoolToolApplication(None)
        self.persons_container = app['persons']
        self.resources_container = app['resources']

        self.courses = {}
        for course in ICourseContainer(self.context).values():
            course_id = course.course_id
            if course_id is not None:
                course_ids = [ID.strip() for ID in course_id.split(",")]
                for ID in course_ids:
                    self.courses[ID] = course

        self.teachers = {}
        group_container = IGroupContainer(self.context)
        teacher_group = group_container.get('teachers')
        if teacher_group:
            for teacher in teacher_group.members:
                demographics = IDemographics(teacher, None)
                if demographics is None:
                    continue
                teacher_id = demographics.get('ID')
                if teacher_id is None:
                    continue
                self.teachers[teacher_id] = teacher

        self.wb = wb
        for self.sheet_name in self.wb.sheet_names():
            if 'SectionDetails' not in self.sheet_name:
                continue
            self.process()

    def process(self):
        current_section_id = None
        sections_by_term = []
        sh = self.sheet

        for row in range(1, sh.nrows):
            if sh.cell_value(rowx=row, colx=0) == '':
                break

            num_errors = len(self.errors)
            course_id = self.getRequiredTextFromCell(sh, row, 0).strip()
            section_index = self.getRequiredTextFromCell(sh, row, 1).strip()
            description = self.getRequiredTextFromCell(sh, row, 2).strip()
            building_id = self.getRequiredTextFromCell(sh, row, 3).strip()
            marking_period = self.getRequiredTextFromCell(sh, row, 5).strip()
            term_description = self.getTextFromCell(sh, row, 6).strip()
            start_period = self.getRequiredTextFromCell(sh, row, 7).strip()
            teacher_id = self.getRequiredTextFromCell(sh, row, 8).strip()
            student_id = self.getRequiredTextFromCell(sh, row, 11).strip()
            if num_errors < len(self.errors):
                continue

            course = self.courses.get(course_id)
            if course is None:
                self.error(row, 1, ERROR_INVALID_COURSE_ID)

            building = self.resources_container.get(building_id)
            if building is None:
                self.error(row, 3, ERROR_INVALID_BUILDING_ID)

            parts = marking_period.split('-')
            if len(parts) != 2:
                self.error(row, 5, ERROR_INVALID_MARKING_PERIODS)
            else:
                first_term = self.context.get(parts[0])
                last_term = self.context.get(parts[1])
                if first_term is None or last_term is None:
                    self.error(row, 5, ERROR_INVALID_MARKING_PERIODS)
                elif first_term.first > last_term.first:
                   self.error(row, 5, ERROR_INVALID_MARKING_PERIODS)
                else:
                    terms = [term for term in self.context.values()
                             if (term.first >= first_term.first and
                                 term.first <= last_term.first)]

            teacher = self.teachers.get(teacher_id)
            if teacher is None:
                self.error(row, 8, ERROR_INVALID_TEACHER_ID)

            student = self.persons_container.get(student_id)
            if student is None:
                self.error(row, 11, ERROR_INVALID_STUDENT_ID)

            if num_errors < len(self.errors):
                continue

            section_id = '%s-%s-%s-%s' % (course_id, building_id,
                                          start_period, section_index)
            if section_id != current_section_id:
                current_section_id = section_id
                sections_by_term = self.getSectionsByTerm(section_id, course,
                                                          description, building,
                                                          terms, start_period,
                                                          term_description,
                                                          section_index)

            for section in sections_by_term:
                if teacher not in section.instructors:
                    section.instructors.add(teacher)
                if student not in section.members:
                    section.members.add(student)

    def getSectionsByTerm(self, section_id, course, description, building,
                          terms, start_period, term_description, section_index):
        sections_by_term = []
        previous_section = None
        terms_part = ''
        if term_description:
            terms_part = ', ' + term_description

        for term in terms:
            if start_period.isdigit():
                period = 'P' + start_period
            else:
                period = start_period
            title = '%s, %s%s, %s' % (description, building.title,
                                      terms_part, period)

            sections_container = ISectionContainer(term)
            if section_id in sections_container:
                section = sections_container[section_id]
            else:
                section = Section(title=title, description=course.description)
                sections_container[section_id] = section
            if course not in section.courses:
                section.courses.add(removeSecurityProxy(course))
            if building not in section.resources:
                section.resources.add(removeSecurityProxy(building))

            if previous_section is not None:
                previous_section.next = section
            previous_section = section
            sections_by_term.append(section)

        return sections_by_term


class VirginiaSectionsMegaImporter(importer.FlourishMegaImporter):

    def nextURL(self):
        url = absoluteURL(ISchoolToolApplication(None), self.request)
        return url + '/manage'

    @property
    def importers(self):
        return [
            SectionsImporter,
            ]

    def update(self):
        if "UPDATE_CANCEL" in self.request:
            self.request.response.redirect(self.nextURL())
            return

        if "UPDATE_SUBMIT" not in self.request:
            return

        xlsfile = self.request.get('xlsfile', '')
        if not xlsfile:
            return
        self.data_provided = True

        wb = xlrd.open_workbook(file_contents=xlsfile.read())

        if wb is None:
            return

        for importer in self.importers:
            imp = importer(self.context, self.request)
            imp.import_data(wb)
            self.errors.extend(imp.errors)

        if not self.errors:
            self.request.response.redirect(self.nextURL())

