#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Virginia report views.
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.i18n.interfaces.locales import ICollator
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zc.table.column import GetterColumn
import zc.resourcelibrary

from schooltool import table
from schooltool.basicperson.interfaces import IDemographics
from schooltool.cando.interfaces import ISectionSkills
from schooltool.cando.interfaces import ISkillsGradebook
from schooltool.cando.skill import querySkillScoreSystem
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.course.interfaces import ISection
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.skin import flourish

from schooltool.virginia import VirginiaMessage as _


class VASectionReportView(flourish.page.Page):

    content_template = ViewPageTemplateFile('templates/va_section_report.pt')
    container_class = 'container extra-wide-container'

    @property
    def section(self):
        gradebook = removeSecurityProxy(self.context)
        return ISection(gradebook)

    @property
    def subtitle(self):
        return self.section.title

    def next_url(self):
        return absoluteURL(self.context, self.request)

    @property
    def scoresystem(self):
        return querySkillScoreSystem()

    @property
    def passing_score(self):
        scoresystem = self.scoresystem
        if self.filter_field_id in self.request:
            rating_score = self.request[self.filter_field_id]
        else:
            rating_score = scoresystem._minPassingScore
        return rating_score

    @property
    def passing_scores(self):
        result = []
        scoresystem = self.scoresystem
        if self.filter_field_id in self.request:
            rating_score = self.request[self.filter_field_id]
        else:
            rating_score = scoresystem._minPassingScore
        for score_info in scoresystem.scores:
            label = score_info[0]
            # XXX: use real scoresystem values here instead of int()
            if int(label) >= int(rating_score):
                result.append(label)
        return '/'.join(reversed(result))

    @property
    def filter_field_id(self):
        return 'minimum_rating_score'

    @property
    def scores(self):
        result = []
        scoresystem = self.scoresystem
        for score_info in scoresystem.scores:
            if self.filter_field_id in self.request:
                selected = score_info[0] == self.request[self.filter_field_id]
            else:
                selected = scoresystem.isPassingScore(score_info[0])
            result.append({
                    'id': score_info[0],
                    'title': score_info[1],
                    'selected': selected,
                    })
        return result
            
    def update(self):
        course_titles = []
        course_government_ids = []
        for course in self.section.courses:
            course_titles.append(course.title)
            course_government_ids.append(course.government_id or '')
        self.course = {
            'title': ', '.join(course_titles),
            'government_id': ', '.join(course_government_ids),
            }
        instructors = []
        for person in self.section.instructors:
            instructors.append(person.title)
        self.instructors = '<br />'.join(instructors)
        rows = []
        collator = ICollator(self.request.locale)
        worksheets = self.context.__parent__.__parent__
        students_count = len(self.section.members)
        for student in sorted(self.section.members,
                              key=lambda x:x.title,
                              cmp=collator.cmp):
            demographics = IDemographics(student)
            info = {
                'student': student.title,
                'student_id': student.username,
                'required_container_id': 'required-%s' % student.username,
                'all_container_id': 'all-%s' % student.username,
                'student_id': demographics.get('ID', ''),
                }
            info.update(self.getStudentSkillsData(student, worksheets))
            rows.append(info)
        self.rows = rows
        attainment_labels = {100: '90-100%',
                             90: '80-89%',
                             80: '70-79%',
                             70: '60-69%',
                             60: '<= 59%'}
        rows_attainment = {100: 0,
                           90: 0,
                           80: 0,
                           70: 0,
                           60: 0}
        rows_attainment80 = {u'80% or above': 0,
                             u'Below 80%': 0}
        for row in self.rows:
            rows_attainment[row['rows_attainment']] += 1
            rows_attainment80[row['rows_attainment80']] += 1
        self.rows_attainment = []
        for key in sorted(rows_attainment, reverse=True):
            info = {"scale": attainment_labels[key],
                    "student_count": rows_attainment[key]}
            try:
                p = rows_attainment[key] / float(students_count)
            except (ZeroDivisionError,):
                p = 0
            info["percentage"] = self.formatPercentage(p)
            self.rows_attainment.append(info)
        self.rows_attainment80 = []
        for key in sorted(rows_attainment80):
            info = {"scale": key,
                    "student_count": rows_attainment80[key]}
            try:
                p = rows_attainment80[key] / float(students_count)
            except (ZeroDivisionError,):
                p = 0
            info["percentage"] = self.formatPercentage(p)
            self.rows_attainment80.append(info)

    def getStudentSkillsData(self, student, worksheets):
        required_total = 0
        required_evaluated = 0
        required_competent = 0
        all_total = 0
        all_evaluated = 0
        all_competent = 0
        for worksheet in worksheets.values():
            gradebook = ISkillsGradebook(worksheet)
            for skill in gradebook.activities:
                all_total += 1
                score = gradebook.getScore(student, skill)
                evaluated = score is not None and score.value is not UNSCORED
                # XXX: use real scoresystem values here instead of int()
                competent = evaluated and \
                            int(score.value) >= int(self.passing_score)
                if evaluated:
                    all_evaluated += 1
                if competent:
                    all_competent += 1
                if skill.required:
                    required_total += 1
                    if evaluated:
                        required_evaluated += 1
                    if competent:
                        required_competent += 1
        try:
            required_rating = required_competent / float(required_evaluated)
        except (ZeroDivisionError,):
            required_rating = 0
        try:
            required_completed = required_competent / float(required_total)
        except (ZeroDivisionError,):
            required_completed = 0
        try:
            all_rating = all_competent / float(all_evaluated)
        except (ZeroDivisionError,):
            all_rating = 0
        if required_completed >= 90:
            rows_attainment = 100
        elif required_completed >= 80:
            rows_attainment = 90
        elif required_completed >= 70:
            rows_attainment = 80
        elif required_completed >= 60:
            rows_attainment = 70
        else:
            rows_attainment = 60
        if required_completed >= 80:
            rows_attainment80 = u'80% or above'
        else:
            rows_attainment80 = u'Below 80%'
        return {
            'required_total': required_total,
            'required_evaluated': required_evaluated,
            'required_competent': required_competent,
            'required_rating': self.formatPercentage(required_rating),
            'required_completed': self.formatPercentage(required_completed),
            'required_data': [required_total, required_evaluated, required_competent],
            'all_total': all_total,
            'all_evaluated': all_evaluated,
            'all_competent': all_competent,
            'all_rating': self.formatPercentage(all_rating),
            'all_data': [all_total, all_evaluated, all_competent],
            'rows_attainment': rows_attainment,
            'rows_attainment80': rows_attainment80,
            }

    def formatPercentage(self, number):
        return '%.0f%%' % (number * 100)


class VASectionChartReportView(VASectionReportView):

    content_template = ViewPageTemplateFile(
        'templates/va_section_report_chart.pt')


class SectionReportView(flourish.page.Page):

    skill_type_filter_id = 'skill-type-filter'
    score_filter_id = 'score-filter'
    cutoff_filter_id = 'cutoff-filter'
    form_container_id = 'section-report-form-container'
    score_colors_container_id = 'score-colors-container'

    @Lazy
    def gradebook(self):
        return removeSecurityProxy(self.context)

    @Lazy
    def section(self):
        return ISection(self.gradebook)

    @property
    def subtitle(self):
        return self.section.title

    @Lazy
    def scoresystem(self):
        return querySkillScoreSystem()

    @Lazy
    def skills(self):
        result = []
        for worksheet in ISectionSkills(self.section).values():
            gradebook = ISkillsGradebook(worksheet)
            for skill in worksheet.values():
                result.append((skill, gradebook))
        return self.filter_skills(result)

    def filter_skills(self, skills):
        skill_type = self.request.get(self.skill_type_filter_id)
        if skill_type is not None and 'required' in skill_type:
            condition = lambda (skill, gradebook): skill.required
        else:
            condition = None
        return filter(condition, skills)

    @property
    def cutoff(self):
        default = 80 # XXX: sync this with filter's value
        result = self.request.get(self.cutoff_filter_id)
        try:
            return int(result)
        except (TypeError, ValueError,):
            return default

    @property
    def passing_score(self):
        scoresystem = self.scoresystem
        if self.score_filter_id in self.request:
            passing_score = self.request[self.score_filter_id]
        else:
            passing_score = scoresystem._minPassingScore
        return passing_score

    def render(self, *args, **kw):
        zc.resourcelibrary.need('schooltool.virginia.flourish')
        return super(SectionReportView, self).render(*args, **kw)


class SectionReportByStudentView(SectionReportView):

    pass


class SectionReportBySkillView(SectionReportView):

    container_class = 'container widecontainer'


class ReportDetailsViewlet(flourish.viewlet.Viewlet):

    template = ViewPageTemplateFile(
        'templates/section_report_details.pt')

    @property
    def course(self):
        course_titles = []
        course_government_ids = []
        section = self.view.section
        for course in section.courses:
            course_titles.append(course.title)
            course_government_ids.append(course.government_id or '')
        return {
            'title': ', '.join(course_titles),
            'government_id': ', '.join(course_government_ids),
            }

    @property
    def instructors(self):
        instructors = []
        section = self.view.section
        for person in section.instructors:
            instructors.append(person.title)
        return '<br />'.join(instructors)


class SectionReportByStudentDescriptionViewlet(flourish.page.Related):

    template = InlineViewPageTemplate('''
      <tal:block i18n:domain="schooltool.virginia">
        <div class="header" i18n:translate="">
          Note:
        </div>
        <div class="body">
          <p i18n:domain="" i18n:translate="">
            Students are expected to achieve a satisfactory rating
            (one of the three highest marks) on the Student Competency
            Record (SCR) rating scale on at least 80% of the required
            (essential) competencies.
          </p>
        </div>
      </tal:block>
    ''')


class SectionReportBySkillDescriptionViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <p i18n:domain="" i18n:translate="">
        XXX This report description goes here XXX
      </p>
    ''')


class SectionReportTableViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <div tal:content="structure view/view/providers/ajax/view/section/charts_table"/>
    ''')


def student_id_getter(student, formatter):
    demographics = IDemographics(student)
    return demographics.get('ID', '')


class SectionReportChartsColumn(GetterColumn):

    template = ViewPageTemplateFile(
        'templates/section_report_charts_column.pt')


class SkillsColumn(SectionReportChartsColumn):

    def __init__(self, view):
        self.view = view
        title = _('% of Skills Below and Above Passing')
        name = 'skills'
        super(SkillsColumn, self).__init__(title, name=name)

    def renderCell(self, item, formatter):
        container_id = 'skills-data-%s' % item.username
        student_data, total = self.studentSkillsData(item)
        data = []
        passing_index = None
        scores = reversed(self.view.scoresystem.scores)
        score_labels = []
        for i, score in enumerate(scores):
            score_labels.append(score[1])
            data.append(student_data[score[0]])
            if score[0] == self.view.passing_score:
                passing_index = i
        # XXX: hack to make the ViewPageTemplateFile to work
        self.context = item
        self.request = formatter.request
        return self.template(total=total,
                             passing_index=passing_index,
                             data=data,
                             container_id=container_id,
                             score_labels = score_labels,
                             cutoff = self.view.cutoff)

    def studentSkillsData(self, student):
        result = {}
        count = 0
        evaluated = 0
        for score in self.view.scoresystem.scores:
            result[score[0]] = 0
        for skill, gradebook in self.view.skills:
            count += 1
            score = gradebook.getScore(student, skill)
            if score is not None and score.value is not UNSCORED:
                evaluated += 1
                result[score.value] += 1
        skill_type = self.view.request.get(self.view.skill_type_filter_id)
        if skill_type is not None and 'evaluated' in skill_type:
            total = evaluated
        else:
            total = count
        return result, total


class StudentsColumn(SectionReportChartsColumn):

    def __init__(self, view):
        self.view = view
        self.total = len(view.section.members)
        title = _('Students (${count})',
                  mapping={'count': self.total})
        name = 'students'
        super(StudentsColumn, self).__init__(title, name=name)

    def renderCell(self, item, formatter):
        skill, gradebook = item
        container_id = 'skills-data-%s' % skill.__name__
        skill_data = self.skillScoresData(skill, gradebook)
        data = []
        passing_index = None
        scores = reversed(self.view.scoresystem.scores)
        score_labels = []
        for i, score in enumerate(scores):
            score_labels.append(score[1])
            data.append(skill_data[score[0]])
            if score[0] == self.view.passing_score:
                passing_index = i
        # XXX: hack to make the ViewPageTemplateFile to work
        self.context = item
        self.request = formatter.request
        return self.template(total=self.total,
                             passing_index=passing_index,
                             data=data,
                             container_id=container_id,
                             score_labels = score_labels,
                             cutoff=self.view.cutoff)

    def skillScoresData(self, skill, gradebook):
        result = {}
        for score in self.view.scoresystem.scores:
            result[score[0]] = 0
        for student in self.view.section.members:
            score = gradebook.getScore(student, skill)
            if score is not None and score.value is not UNSCORED:
                result[score.value] += 1
        return result


class SectionReportSortFormatter(table.ajax.AJAXFormSortFormatter):

    script_name = 'ST.virginia.on_section_report_sort'


class SectionReportChartsTable(table.ajax.Table):

    batch_size = 0
    table_formatter = SectionReportSortFormatter
    css_classes = ''

    @property
    def html_id(self):
        return self.view.form_container_id

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': self.css_classes})


class ByStudentChartsTable(SectionReportChartsTable):

    css_classes = 'section-report section-report-by-student'

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u'Student'),
            getter=lambda i, f: i.title,
            subsort=True)
        ID = table.column.LocaleAwareGetterColumn(
            name='ID',
            title=_(u'Student ID #'),
            getter=student_id_getter,
            subsort=True)
        skills = SkillsColumn(self.view)
        return [title, ID, skills]

    def sortOn(self):
        return (('title', False),)

    def items(self):
        return self.context.members


class BySkillChartsTable(SectionReportChartsTable):

    css_classes = 'section-report section-report-by-skill'

    def columns(self):
        # i => (skill, gradebook)
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i[0].label,
            subsort=True)
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u'Skill'),
            getter=lambda i, f: i[0].title,
            subsort=True)
        students = StudentsColumn(self.view)
        return [label, title, students]

    def sortOn(self):
        return (('label', False),)

    def items(self):
        return self.view.skills


class ScoreFilterViewlet(flourish.page.RefineLinksViewlet): pass


class ScoreFilterMenuViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <select tal:attributes="id view/view/score_filter_id;
                              name view/view/score_filter_id;"
              class="select-widget report-filter">
        <option tal:repeat="score view/scores"
                tal:content="score/title"
                tal:attributes="value score/id;
                                selected score/selected;"
                />
      </select>
    ''')

    def scores(self):
        result = []
        scoresystem = self.view.scoresystem
        score_filter_id = self.view.score_filter_id
        for score_info in scoresystem.scores:
            if score_filter_id in self.request:
                selected = score_info[0] == self.request[score_filter_id]
            else:
                selected = score_info[0] == scoresystem._minPassingScore
            result.append({
                    'id': score_info[0],
                    'title': score_info[1],
                    'selected': selected,
                    })
        return result


class SkillTypeFilterViewlet(flourish.page.RefineLinksViewlet): pass


class SkillTypeFilterOptionsViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <ul class="filter">
        <li tal:repeat="option view/type_options">
          <input type="radio" class="report-filter"
                 tal:attributes="id option/id;
                                 name view/view/skill_type_filter_id;
                                 value option/id;
                                 checked option/checked;"
                 />
          <label tal:attributes="for option/id"
                 tal:content="option/title" />
        </li>
      </ul>
    ''')

    @property
    def options(self):
        return [
            ('all', _('All'), True),
            ('required', _('Required'), False),
            ('evaluated', _('Evaluated'), False),
            ]

    def type_options(self):
        result = []
        filter_id = self.view.skill_type_filter_id
        for prefix, title, default in self.options:
            option_id = '%s.%s' % (prefix, filter_id)
            if filter_id in self.request:
                checked = self.request.get(filter_id) == option_id
            else:
                checked = default
            result.append({
                'id': option_id,
                'title': title,
                'checked': checked,
                })
        return result


class CutoffFilterViewlet(flourish.page.RefineLinksViewlet): pass


class CutoffFilterMenuViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <select tal:attributes="id view/view/cutoff_filter_id;
                              name view/view/cutoff_filter_id;"
              class="select-widget report-filter">
        <option tal:repeat="value view/values"
                tal:content="value/title"
                tal:attributes="value value/id;
                                selected value/selected;"
                />
      </select>
    ''')

    def values(self):
        result = []
        start = 10
        end = 100
        step = 10
        default = 80
        cutoff_filter_id = self.view.cutoff_filter_id
        for value in range(start, end, step):
            if cutoff_filter_id in self.request:
                selected = str(value) == self.request[cutoff_filter_id]
            else:
                selected = value == default
            result.append({
                    'id': value,
                    'title': '%d%%' % value,
                    'selected': selected,
                    })
        return result


class SectionReportDoneButtonViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <h3 i18n:domain="schooltool.virginia">
        <a tal:attributes="href view/nextURL" i18n:translate="">
          Done
        </a>
      </h3>
    ''')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class ScoreColorsViewlet(flourish.page.Related):

    template = ViewPageTemplateFile(
        'templates/section_report_score_colors.pt')


class ScoreColorColumn(GetterColumn):

    template = ViewPageTemplateFile(
        'templates/section_report_score_colors_column.pt')

    def renderCell(self, item, formatter):
        container_id = 'score-color-%s' % item['id']
        # XXX: hack to make the ViewPageTemplateFile to work
        self.context = item
        self.request = formatter.request
        return self.template(container_id=container_id, score=item)


class ScoreColorsTable(table.ajax.Table):

    batch_size = 0

    @property
    def html_id(self):
        return self.view.score_colors_container_id

    def columns(self):
        color = ScoreColorColumn(
            name='color',
            title=_(u'Score Colors'))
        return [color]

    def items(self):
        scores = self.view.scoresystem.scores
        passing = []
        not_passing = []
        for score in scores:
            is_passing = int(score[0]) >= int(self.view.passing_score)
            if is_passing:
                l = passing
            else:
                l = not_passing
            l.append({
                    'id': score[0],
                    'title': score[1],
                    'is_passing': is_passing
                    })
        for i, score in enumerate(passing):
            if len(passing) > 1:
                score['color_weight'] = float(i)/(len(passing)-1)
            else:
                score['color_weight'] = float(i)/len(passing)              
        for i, score in enumerate(reversed(not_passing)):
            if len(not_passing) > 1:
                score['color_weight'] = float(i)/(len(not_passing)-1)
            else:
                score['color_weight'] = float(i)/len(not_passing)
        return passing + not_passing

    def sortOn(self):
        return None

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'score-colors-table'})
