#!/usr/bin/python

import sys
import urllib
import json
import cPickle as pickle
import xlwt

VERSO_URL = 'http://api.cteresource.org'


class DB(dict):

    filename = None
    _changed = 1
    _valid = False
    _savepoint_every = 1000

    def __init__(self, filename):
        dict.__init__(self)
        self.filename = filename

    def __setitem__(self, *args, **kw):
        self._changed += 1
        if self._changed % self._savepoint_every == 0:
            self.backup()
            self.save()
        dict.__setitem__(self, *args, **kw)

    def load(self, quiet=False):
        print 'Loading...'
        self.clear()
        try:
            f = open(self.filename)
        except IOError:
            if not quiet:
                print 'Failed to load', self.filename
            return False
        try:
            data = pickle.load(f)
        except EOFError:
            print 'Failed to load', self.filename
            return False
        self.update(data)
        f.close()
        self._valid = True
        return True

    def backup(self):
        if not self._valid:
            return
        try:
            db = open(self.filename)
        except IOError:
            return
        backup_name = self.filename+'.bak'
        print 'Making backup', backup_name
        backup = open(backup_name, 'w')
        backup.write(db.read())
        db.close()
        backup.close()

    def save(self, filename=None):
        if filename is None:
            filename = self.filename
        f = open(filename, 'w')
        print 'Saving', filename
        pickle.dump(dict(self), f)
        f.close()
        self._valid = True


LOADED = DB('verso-db')


class NotFound(Exception):
    pass


def retrieve(url):
    while url.startswith('/'):
        url = url[1:]
    url = VERSO_URL+'/'+url
    f = urllib.urlopen(url)
    data = json.load(f)
    f.close()
    if 'errors' in data:
        if data['errors'] == 'Not Found':
            raise NotFound(url)
        raise Exception(data['errors'], url)
    return data


class Doodad(object):

    def __init__(self, **kw):
        self.raw_data = {}
        if kw:
            self.update(kw)

    @classmethod
    def query(cls, *args, **kw):
        doodad = cls(*args, **kw)
        return LOADED.get(doodad.unique_key, doodad)


    @property
    def is_loaded(self):
        return self.unique_key in LOADED

    @property
    def unique_key(self):
        return self.__class__.__name__, self.id

    def __getattr__(self, name):
        if name.startswith('__'):
            return object.__getattr__(self, name)
        if not hasattr(self, 'id'):
            raise Exception('Cannot load object without id')
        global LOADED
        key = self.unique_key
        if key not in LOADED:
            self.load()
        if key not in LOADED:
            raise AttributeError(name)
        return getattr(LOADED[key], name)

    def load(self):
        key = self.unique_key
        print 'LOADING', key
        LOADED[key] = self
        cn = self.__class__.__name__
        if cn not in BY_CLASS:
            BY_CLASS[cn] = {}
        BY_CLASS[cn][key] = self

    def update(self, data, soft=True):
        if soft:
            raw = self.raw_data
            for k, v in data.items():
                if k not in raw:
                    raw[k] = v
        else:
            self.raw_data.update(data)
        for key in self.raw_data:
            try:
                setattr(self, key, self.raw_data[key])
            except:
                pass # meh

    def __repr__(self):
        return '<%s %s %r>' % (self.__class__.__name__,
                               self.id,
                               self.title)


class Doodads(dict):
    _factory = Doodad
    def __init__(self, *data):
        dict.__init__(self)
        self.update(data)

    def update(self, raw_items):
        factory = self._factory
        for raw in raw_items:
            item = factory(**raw)
            self[item.id] = item

    def load(self, preload_children=False):
        if preload_children:
            for c in self.values():
                key = c.unique_key
                if key not in LOADED:
                    try:
                        c.load()
                    except NotFound:
                        print 'Failed to preload', key
                        pass

    def __repr__(self):
        return '<%s (%s %s(s))>: %s' % (
            self.__class__.__name__, len(self), self._factory.__name__, self.keys())


class CourseBoundMixin(object):

    @property
    def unique_key(self):
        return self.__class__.__name__, self.course_code, self.course_edition, self.id

    @property
    def course(self):
        if not hasattr(self, 'course_code') or not hasattr(self, 'course_edition'):
            raise AttributeError('course')
        return Course(code=self.course_code, edition=self.course_edition)


class Task(CourseBoundMixin, Doodad):


    def load(self):
        course = self.course
        data = retrieve('/courses/%s,%s/tasks/%s' %
                        (course.code, course.edition, self.id))
        self.update(data)
        super(Task, self).load()

    def __repr__(self):
        return '<%s %s %s %r>' % (self.__class__.__name__,
                               self.id,
                               self.essential and 'required' or 'optional',
                               self.statement)


class TaskList(CourseBoundMixin, Doodads):
    _factory = None

    def __init__(self, course_code, course_edition, data=()):
        self.course_code = course_code
        self.course_edition = course_edition
        Doodads.__init__(self, *data)

    @property
    def _factory(self):
        def CourseTask(*a, **kw):
            kwe = dict(kw)
            kwe['course_code'] = self.course_code
            kwe['course_edition'] = self.course_edition
            return Task(*a, **kwe)
        return CourseTask


class DutyArea(CourseBoundMixin, Doodad):

    def update(self, data, soft=True):
        Doodad.update(self, data, soft=soft)
        if 'tasks' in data:
            self.tasks = TaskList(self.course_code, self.course_edition, data=self.raw_data['tasks'])

    def load(self, **kw):
        self.tasks = TaskList(self.course_code, self.course_edition, data=self.raw_data['tasks'])
        super(DutyArea, self).load(**kw)


class DutyAreas(CourseBoundMixin, Doodads):
    _factory = None

    def __init__(self, course_code, course_edition, data=()):
        self.course_code = course_code
        self.course_edition = course_edition
        Doodads.__init__(self, *data)

    @property
    def _factory(self):
        def CourseDutyArea(*a, **kw):
            kwe = dict(kw)
            kwe['course_code'] = self.course_code
            kwe['course_edition'] = self.course_edition
            kwe['id'] = len(self)
            return DutyArea(*a, **kwe)
        return CourseDutyArea

    def load(self, **kw):
        self.clear()
        course = self.course
        data = retrieve('courses/%s,%s/tasks/' % course.id)
        self.update(data['duty_areas'])
        super(DutyAreas, self).load(**kw)


class Course(Doodad):

    @property
    def unique_key(self):
        return self.__class__.__name__, self.code, self.edition

    @property
    def id(self):
        return self.code, self.edition

    def __getattr__(self, name):
        if name.startswith('__'):
            return object.__getattr__(self, name)
        if not hasattr(self, 'code') or not hasattr(self, 'edition'):
            raise Exception('Cannot load course without code and edition')
        return super(Course, self).__getattr__(name)

    def load(self, **kw):
        data = retrieve('courses/%s,%s' % self.id)
        self.update(data)
        self.occupations = []
        for occupations in self.raw_data['occupation_data']:
            for occupation in occupations['occupations']:
                self.occupations.append(Occupation(**occupation))
        try:
            duty_data = retrieve('courses/%s,%s/tasks/' % self.id)
            self.duty_areas = DutyAreas(self.code, self.edition, data=duty_data['duty_areas'])
        except NotFound:
            self.duty_areas = DutyAreas(self.code, self.edition)
        super(Course, self).load(**kw)


class CourseList(Doodads):
    _factory = Course

    def load(self, **kw):
        data = retrieve('courses/')
        self.update(data['courses'])
        super(CourseList, self).load(**kw)


class Occupation(Doodad):

    def load(self, **kw):
        data = retrieve('occupations/%s' % self.id)
        self.update(data['occupation'])
        super(Occupation, self).load(**kw)


class OccupationList(Doodads):
    _factory = Occupation

    def load(self, **kw):
        data = retrieve('occupations/')

        for entry in data['occupation_data']:
            self.update(entry['occupations'])
        super(OccupationList, self).load(**kw)


class Pathway(Doodad):

    def load(self, **kw):
        data = retrieve('pathways/%s' % self.id)
        self.update(data['pathway'])
        self.occupations = OccupationList(*self.raw_data['occupations'])
        super(Pathway, self).load(**kw)


class PathwayList(Doodads):
    _factory = Pathway

    def load(self, **kw):
        data = retrieve('pathways/')
        self.update(data['pathways'])
        super(PathwayList, self).load(**kw)


class Cluster(Doodad):

    def load(self, **kw):
        data = retrieve('clusters/%s' % self.id)
        self.update(data['cluster'])
        self.pathways = PathwayList(*self.raw_data['pathways'])
        super(Cluster, self).load(**kw)


class ClusterList(Doodads):
    _factory = Cluster

    def load(self, **kw):
        data = retrieve('clusters/')
        self.update(data['clusters'])
        super(ClusterList, self).load(**kw)


BY_CLASS = dict()
def update_by_class():
    for k, v in LOADED.items():
        key = v.unique_key
        cn = v.__class__.__name__
        if cn not in BY_CLASS:
            BY_CLASS[cn] = {}
        BY_CLASS[cn][key] = v


def build_xls():
    wb_file = open('verso_skills.xls', 'w')
    wb = xlwt.Workbook()

    YELLOW = 5
    font = xlwt.Font()
    font.bold = True
    header_style = xlwt.XFStyle()
    header_style.font = font
    pattern = xlwt.Pattern()
    pattern.pattern = xlwt.Pattern.SOLID_PATTERN
    pattern.pattern_fore_colour = YELLOW
    header_style.pattern = pattern

    cell_style = xlwt.XFStyle()

    ################################
    ws = wb.add_sheet("Layers")
    ws.write(0, 0, 'ID', header_style)
    ws.write(0, 1, 'Title', header_style)
    ws.write(0, 2, 'Parents', header_style)

    items = [
        ('skill', 'Skill'),
        ('skillset', 'Skill Set'),
        ('course', 'Course'),
        ('cluster', 'Cluster'),
        ]
    for index, (name, title) in enumerate(items):
        ws.write(index + 1, 0, name, cell_style)
        ws.write(index + 1, 1, title, cell_style)
        if index < len(items) - 1:
            ws.write(index + 1, 2, items[index + 1][0], cell_style)

    ################################
    ws = wb.add_sheet("Documents")
    ws.write(0, 0, 'ID', header_style)
    ws.write(0, 1, 'Title', header_style)
    ws.write(0, 2, 'Description', header_style)
    ws.write(0, 3, 'Hierarchy', header_style)

    ws.write(1, 0, 'cte', cell_style)
    ws.write(1, 1, 'Virginia CTE', cell_style)
    hierarchy = ', '.join([name for name, title in items])
    ws.write(1, 3, hierarchy, cell_style)

    ################################
    nodes_ws = wb.add_sheet("Nodes")
    nodes_ws.write(0, 0, 'ID', header_style)
    nodes_ws.write(0, 1, 'Title', header_style)
    nodes_ws.write(0, 2, 'Description', header_style)
    nodes_ws.write(0, 3, 'Label', header_style)
    nodes_ws.write(0, 4, 'Documents', header_style)
    nodes_ws.write(0, 5, 'Parents', header_style)
    nodes_ws.write(0, 6, 'Layers', header_style)
    nodes_ws.write(0, 7, 'SkillSets', header_style)

    skillsets_ws = wb.add_sheet("SkillSets")
    skillsets_ws.write(0, 0, 'ID', header_style)
    skillsets_ws.write(0, 1, 'Title', header_style)
    skillsets_ws.write(0, 2, 'External ID', header_style)
    skillsets_ws.write(0, 3, 'Label', header_style)

    skills_ws = wb.add_sheet("Skills")
    skills_ws.write(0, 0, 'SkillSet ID', header_style)
    skills_ws.write(0, 1, 'Skill ID', header_style)
    skills_ws.write(0, 2, 'Title', header_style)
    skills_ws.write(0, 3, 'Equivalent', header_style)
    skills_ws.write(0, 4, 'Description', header_style)
    skills_ws.write(0, 5, 'External ID', header_style)
    skills_ws.write(0, 6, 'Label', header_style)
    skills_ws.write(0, 7, 'Required', header_style)
    skills_ws.write(0, 8, 'Retired', header_style)

    node_row = skillset_row = skill_row = 1

    occupation_dict = {}
    for cluster in BY_CLASS['Cluster'].values():
        cluster_id = 'cluster-' + unicode(cluster.id)
        nodes_ws.write(node_row, 0, cluster_id, cell_style)
        nodes_ws.write(node_row, 1, unicode(cluster.title), cell_style)
        nodes_ws.write(node_row, 4, 'cte', cell_style)
        nodes_ws.write(node_row, 6, 'cluster', cell_style)
        node_row += 1

        for pathway in cluster.pathways.values():
            for occupation in pathway.occupations.values():
                occupation_dict.setdefault(occupation.id, set()).add(cluster)

    course_dict = {}
    for course in BY_CLASS['Course'].values():
        name = '%s-%s' % (course.code, course.edition)
        course_dict[name] = course

    for name in sorted(course_dict):
        course = course_dict[name]

        nodes_ws.write(node_row, 0, name, cell_style)
        nodes_ws.write(node_row, 1, course.title, cell_style)
        nodes_ws.write(node_row, 3, course.code, cell_style)
        clusters = set()
        for occupation in course.occupations:
            for cluster in occupation_dict.get(occupation.id, []):
                clusters.add(cluster)
        parents_list = ['cluster-' + unicode(cluster.id)
                        for cluster in clusters]
        nodes_ws.write(node_row, 5, ', '.join(parents_list), cell_style)
        nodes_ws.write(node_row, 6, 'course', cell_style)
        skillsets_list = [name + '-' + unicode(duty_area.id)
                          for duty_area in course.duty_areas.values()]
        nodes_ws.write(node_row, 7, ', '.join(skillsets_list), cell_style)
        node_row += 1

        task_count = 0
        for duty_area_index, duty_area in enumerate(course.duty_areas.values()):
            label = '%02d' % (duty_area_index + 1)
            skillset_id = name + '-' + unicode(duty_area.id)
            skillsets_ws.write(skillset_row, 0, skillset_id, cell_style)
            skillsets_ws.write(skillset_row, 1, unicode(duty_area.title), cell_style)
            skillsets_ws.write(skillset_row, 3, label, cell_style)
            skillset_row += 1

            for task_index, task in enumerate(duty_area.tasks.values()):
                task_count += 1
                label = '%03d' % task_count
                external_id = 'http://competencies.cteresource.org/' + task.id
                skills_ws.write(skill_row, 0, skillset_id, cell_style)
                skills_ws.write(skill_row, 1, label, cell_style)
                skills_ws.write(skill_row, 2, unicode(task.statement), cell_style)
                skills_ws.write(skill_row, 4, unicode(task.definition), cell_style)
                skills_ws.write(skill_row, 5, external_id, cell_style)
                skills_ws.write(skill_row, 6, label, cell_style)
                skills_ws.write(skill_row, 7, unicode(task.essential), cell_style)
                skill_row += 1

    wb.save(wb_file)


def loadAll():
    cl = ClusterList()
    cl.load(preload_children=True)
    pl = PathwayList()
    pl.load(preload_children=True)
    ol = OccupationList()
    ol.load(preload_children=True)
    cl = CourseList()
    cl.load(preload_children=True)
    update_by_class()
    for course in BY_CLASS['Course'].values():
        for duty_area in course.duty_areas.values():
            if not duty_area.is_loaded:
                duty_area.load()
            for task in duty_area.tasks.values():
                if not task.is_loaded:
                    task.load()

    LOADED.save(filename=None)
    LOADED.backup()

    print 'AFTER loading'
    print 'DB entries:', len(LOADED)
    print 'Classes:\n'+'\n'.join([str((k, len(v))) for k, v in BY_CLASS.items()])


def printDBSummary():
    print 'DB entries:', len(LOADED)
    print 'Classes:\n'+'\n'.join([str((k, len(v))) for k, v in BY_CLASS.items()])
    print '-'*80
    print ''

    all_ids = set([t.id for t in BY_CLASS['Task'].values()])
    print 'unique tasks (by ID):', len(all_ids)
    all_ids_ext = set([(t.id, t.essential) for t in BY_CLASS['Task'].values()])
    print 'unique assigned tasks (by ID+essential):', len(all_ids_ext)


    by_course = {}
    for task in BY_CLASS['Task'].values():
        i = task.course_edition, task.course_code
        if i not in by_course:
            by_course[i] = 0
        by_course[i] += 1

    print 'Tasks per course:'
    task_counts = [i for i in by_course.values() if i > 0]
    t_min = min(task_counts)
    print 'min', t_min
    t_max = max(task_counts)
    print 'max', t_max

    print 'avg', sum(task_counts) / float(len(task_counts))

    print 'taskless courses (ignored):', len(by_course) - len(task_counts)

    for k, i in sorted(by_course.items()):
        if i == t_min:
            print 'Min tasks ('+str(i)+')', Course(edition=k[0], code=k[1])
        if i == t_max:
            print 'Max tasks ('+str(i)+')', Course(edition=k[0], code=k[1])

    print '-'*80
    print ''

    # Overview.
    #
    # Layers: Cluster -> Pathway -> Occupation -> Course
    #
    # DutyArea == SkillSet
    # Task     == Skill

    def print_header(text):
        print text
        print '='*len(text)
    

    # Example: list clusters

    clusters = BY_CLASS['Cluster'].values()
    print_header('Available clusters')
    for c in clusters:
        print ' *', c.title
    print '\n'

    pattrs = lambda i: [n for n in dir(i) if not n.startswith('_')]

    # Let's pick a course
    courses = BY_CLASS['Course'].values()
    course = courses[0]

    # And now load it in another way (courses are exceptional in
    # a way that they need both code and edition, while other
    # objects need only id)
    code = course.code
    edition = course.edition
    course = Course.query(code=code, edition=edition)

    print_header('Example course')
    print course


    print '-- Taught for'
    # There can be many occupations here:
    for data in course.occupation_data:
        cluster = Cluster.query(id=data['cluster']['id'])
        pathway = Pathway.query(id=data['pathway']['id'])
        print '   ', cluster
        print '   ', pathway
        print '    Occupations:'
        for occup in data['occupations']:
            occupation = Occupation.query(id=occup['id'])
            print '      ', occupation
        print '    ------'

    print '-- Skills:'

    for duty_area in course.duty_areas.values():
        #duty_area is a skillset
        print '  ', duty_area.title
        for task in duty_area.tasks.values():
            print '    -', task.statement, '(%s)' % task.id

    print '\n\n'


    print "Let's look at other objects in more detail.\n"

    print_header('Example cluster (node)')
    print cluster, 'ID:', cluster.id
    print cluster.title
    print cluster.description
    print cluster.pathways
    print 'All attributes:', pattrs(cluster)
    print ''

    pathway =cluster.pathways.values()[0]
    print_header('Example pathway (node)')
    print pathway, 'ID:', pathway.id
    print pathway.title
    print 'All attributes:', pattrs(pathway)
    print ''

    print_header('Example occupation (node)')
    print occupation, 'ID:', occupation.id
    print occupation.title
    print occupation.description
    print 'All attributes:', pattrs(occupation)
    print ''


    print_header('Example course (node)')
    print course, 'code:', course.code, 'edition:', course.edition
    print course.title
    print course.description.strip()[:33] + '<...>' # long description
    print 'All attributes:', pattrs(course)
    print 'Related courses:\n   ', '\n    '.join([
        Course.query(code=related['code'], edition=course.edition).title
        for related in course.related_courses])
    print ''

    print_header('Example duty area (skillset)')
    print duty_area, 'ID:', duty_area.id
    print duty_area.title
    print duty_area.course
    print 'All attributes:', pattrs(duty_area)
    print ''

    print_header('Example task (skill)')
    print task, 'ID:', task.id

    print task.statement # title
    # These are very long:
    print task.definition.strip()[:33] + '<...>' # description
    print task.questions.strip()[:33] + '<...>' # even more description
    # essential means required/optional for this course and skillset
    print task.essential and 'required' or 'optional', 'in', task.course
    print 'All attributes:', pattrs(task)
    print ''


if __name__ == '__main__':
    download_flag = True
    if len(sys.argv) > 1:
        if sys.argv[1] == '--no_download':
            download_flag = False
    if download_flag:
        LOADED.load(quiet=True) # pre-load from pickle if available
        loadAll()
    else:
        LOADED.load()
        update_by_class()
    build_xls()

